import 'dart:convert';

import 'package:easy_pass_common/app_config.dart';
import 'package:easy_pass_common/dto/get_request_base.dart';
import 'package:easy_pass_common/dto/index_request_base.dart';
import 'package:easy_pass_common/dto/request_base.dart';
import 'package:easy_pass_common/dto/update_request_base.dart';
import 'package:easy_pass_common/exceptions/unauthorized_exception.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart';
import 'package:get_it/get_it.dart';
import 'package:http/http.dart';

typedef T Preprocessor<T>();

abstract class ServiceBase {
  final Client client;

  ServiceBase(Client client) : this.client = client;

  Future<List<Map<String, Object>>> index(IndexRequestBase request) async {
    var response = await client.get(
      Uri.http(GetIt.instance<AppConfig>().apiBaseUrl, request.path, _processParameters(request.parameters)), 
      headers: await _buildDefaultHeaders()
    );

    return _processPagedResponse(response);
  }

  Future<Map<String, Object>> post(RequestBase request) async {
    var response = await client.post(
      Uri.http(GetIt.instance<AppConfig>().apiBaseUrl, request.path),
      body: json.encode(request.parameters),
      headers: await _buildDefaultHeaders()
    );

    return _processResponse(response);
  }

  Future<Map<String, Object>> geti(GetRequestBase request) async {
    var response = await client.get(
      Uri.http(GetIt.instance<AppConfig>().apiBaseUrl, '${request.path}/${request.id}'),
      headers: await _buildDefaultHeaders()
    );

    return _processResponse(response);    
  }

  Future<Map<String, Object>> update(UpdateRequestBase request) async {
    var response = await client.put(
      Uri.http(GetIt.instance<AppConfig>().apiBaseUrl, '${request.path}/${request.id}'),
      body: json.encode(request.parameters),
      headers: await _buildDefaultHeaders()
    );

    return _processResponse(response);    
  }  

  Future<Map<String, String>> _buildDefaultHeaders() async {
    var storage = GetIt.instance<FlutterSecureStorage>();

    var jwt = await storage.read(key: 'jwt');  
        
    return {
      "Content-Type" : "application/json",
      "Authorization": "Bearer $jwt"
    };
  }

  Map<String, Object> _processResponse(Response response) {
    return this._processResponseGeneric<Map<String, Object>>(response, () => json.decode(response.body));
  }

  List<Map<String, Object>> _processPagedResponse(Response response) {
    return this._processResponseGeneric<List<Map<String, Object>>>(response, (){
      final List<dynamic> dynamicList = json.decode(response.body);
      final jsonObjectList = List<Map<String, Object>> ();
      dynamicList.forEach((dynamic object) {
        jsonObjectList.add(object as Map<String, Object>);
      });
      return jsonObjectList;
    });
  }

  T _processResponseGeneric<T>(Response response, Preprocessor<T> preprocessor) {
    switch (response.statusCode) {
      case 200:
      case 201:
        return preprocessor();
      case 401:
        throw UnAuthorizedException();
      default:
        throw Exception();
    }    
  }
  
  Map<String, String> _processParameters(Map<String, Object> parameters) {
    if (parameters == null || parameters.isEmpty) {
      return Map<String, String>();
    }
    final res = Map<String, String>();
    parameters.forEach((String key, Object value){
      res[key] = value.toString();
    });

    return res;
  }
}